function [ freq, val ] = reco_cardI_basic1( fct_handle, d, N, dims, threshold )
%RECO_CARDI_BASIC1 Utility function for the function sfft_algorithm.
%   This function realizes step 1 and 2a of Algorithm 1/2.
%   The one-dimensionally projected Fourier coefficients, which are
%   projected are computed to the component specificied by the parameter dims,
%   and the projected Fourier coefficients larger than the relative threshold
%   are returned with the corresponding frequencies.
%
%   Input parameters:
%      fct_handle - function handle (block box) which accepts sampling
%                   nodes as an array of size [num_nodes,d]
%      d - dimensionality
%      N - parameter for search space Gamma = [-N,...,N]^d of possible
%          frequencies
%      threshold - (optional) relative threshold theta used for truncation
%      dims - the component to project to
%      threshold - relative threshold used for truncation.
%
%    Ouput parameters:
%       freq - frequencies belonging to the projected Fourier coefficients
%              larger than the relative threshold
%       val - projected Fourier coefficients larger than the relative threshold
%
% Copyright (c) 2015 Toni Volkmer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

if d < 1
  error('d must be at least 1');
end

if ((d > 3) && (nargin < 4))
  error('if d is larger than 3, dims have to be specified');
end

if nargin > 3
  dims = unique(dims);
  if min(dims) < 1
    error('values in dims must be at least 1');
  end
  if max(dims) > d
    error('values in dims must not be greater than d');
  end
else
  if d == 1
    dims = 1;
  elseif d == 2
    dims = [ 1 2 ];
  else
    dims = [ 1 2 3 ];
  end
end

if nargin < 5
  threshold = 1e-12;
end

if length(dims) == 1
  freq_cand = (-N:N)';
  x = freq_cand / (2*N+1);
elseif length(dims) == 2
  [K,L] = meshgrid(-N:N,-N:N);
  freq_cand = [reshape(K,(2*N+1)^2,1) reshape(L,(2*N+1)^2,1)];
  clear K L;
  x = freq_cand ./ (2*N+1);
elseif length(dims) == 3
  [K,L,M] = meshgrid(-N:N,-N:N,-N:N);
  freq_cand = [reshape(M,(2*N+1)^3,1) reshape(K,(2*N+1)^3,1)  reshape(L,(2*N+1)^3,1)];
  clear K L M;
  x = freq_cand ./ (2*N+1);
else
  error('unsupported');
end

newx = repmat(rand(1,d),size(x,1),1);
for t = 1:length(dims)
  newx(:,dims(t)) = x(:,t);
end
x = mod(newx,1);
clear newx;
samples = fct_handle(x);
clear x;

if length(dims) == 1
  val_all = fftshift(fft(ifftshift(samples))) / (2*N+1);
elseif length(dims) == 2
  samples2 = reshape(samples, 2*N+1, 2*N+1);
  clear samples;
  val_all2 = fftshift(fft2(ifftshift(samples2))) / (2*N+1)^2;
  clear samples2;
  val_all = reshape(val_all2,(2*N+1)^2,1);
  clear vall_all2;
else
  samples2 = reshape(samples, 2*N+1, 2*N+1, 2*N+1);
  clear samples;
  val_all2 = fftshift(fftn(ifftshift(samples2))) / (2*N+1)^3;
  clear samples2;
  val_all = reshape(val_all2,(2*N+1)^3,1);
  clear vall_all2;
end

max_abs_val_all = max(abs(val_all));
absolute_threshold = threshold*max_abs_val_all;
ind = find(abs(val_all) >= absolute_threshold);
freq = freq_cand(ind,:);
val = val_all(ind);

end

