function [ freq, val, freq_cand_sizes, lattice_sizes, sample_sizes, z, run_times ] = sfft_algorithm( fct_handle, d, N, algorithm_name, sparsity_s, sparsity_s_local, threshold_theta, threshold_theta_basic, niter_r, niter_r_basic )
%SFFT_ALGORITHM Sparse high-dimensional FFT based on rank-1 lattice sampling.
%This is an approximate algorithm for the reconstruction of sparse
%high-dimensional trigonometric polynomials, which adaptively constructs the
%index set of frequencies belonging to the non-zero Fourier coefficients in a
%dimension incremental way. For further details, please read
%
%   D. Potts and T. Volkmer
%   Sparse high-dimensional FFT based on rank-1 lattice sampling
%   Submitted, 2015
%
%   Input parameters:
%      fct_handle - function handle (block box) which accepts sampling
%                   nodes as an array of size [num_nodes,d]
%      d - dimensionality
%      N - parameter for search space Gamma = [-N,...,N]^d of possible
%          frequencies
%      algorithm_name - (optional) type of algorithm,
%                       either 'a1r1l' (default) for Algorithm 1
%                       or 'a2r1l' for Algorithm 2
%      sparsity_s - (optional) target sparsity of output, default (2*N+1)^d
%      sparsity_s_local - (optional) value of sparsity_s in local loops,
%                         uses the value of sparsity_s as default
%      threshold_theta - (optional) relative threshold theta used for
%                        truncation in step 2e, default is 1e-12
%      threshold_theta_basic - (optional) relative threshold theta used for
%                              truncation in steps 1 and 2a,
%                              uses the value of threshold_theta as default 
%      niter_r - (optional) number of detection iteration r in steps 2b-2e,
%                default is 1
%      niter_r_basic - (optional) number of detection iteration r in
%                      steps 1 and 2a, uses the value of niter_r as default
%
%   Output parameters:
%      freq - detected frequencies belonging to the (approximately)
%             sparsity_s largest Fourier coefficients, size [sparsity_s,d]
%      val - corresponding Fourier coefficients
%      freq_cand_sizes - array of length d with number of investigated
%                        frequencies in each dimension increment step t
%      lattice_sizes - array of length d with size M of reconstructing
%                      rank-1 lattices used for the sampling
%                      in each dimension increment step t
%      sample_sizes - array of length d with number of samples taken from
%                     the function fct_handle in each dimension increment
%                     step t
%      z - generating vector z of reconstructing rank-1 lattice for
%          frequencies in freq, lattice_sizes(end) contains the
%          corresponding rank-1 lattice size M
%      run_times - array of length 5 with runtimes measured for various
%                  parts of the algorithm,
%                  run_times(1): steps 1 and 2a,
%                  run_times(2): building of reconstructing rank-1 lattices
%                  run_times(3): sampling step 2c
%                  run_times(4): inverse rank-1 lattice FFT step 2d
%                  run_times(5): total runtime of the algorithm
%
% Copyright (c) 2015-2016 Toni Volkmer, 2016 Lutz Kaemmerer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

if exist('maxNumCompThreads') == 5
  warning('off', 'MATLAB:maxNumCompThreads:Deprecated');
  nthreads = maxNumCompThreads;
else
  nthreads = -1;
end

if nargin < 4
  algorithm_name = 'a1r1l';
end

if nargin < 5
  sparsity_s = (2*N+1)^d;
end

if nargin < 6
  sparsity_s_local = sparsity_s;
end

if nargin < 7
  threshold_theta = 1e-12;
end

if nargin < 8
  threshold_theta_basic = threshold_theta;
end

if nargin < 9
  niter_r = 1;
end

if nargin < 10
  niter_r_basic = niter_r;
end

check_naturalnumber_or_error(d, 'dimensionality d');
check_naturalnumber_or_error(N, 'parameter N');

switch algorithm_name
  case {'a1r1l','a2r1l','a1r1lsymhc'}
  otherwise
    error('unsupported algorithm: %s', algorithm_name);
end

check_naturalnumber_or_error(sparsity_s, 'parameter sparsity_s');
check_naturalnumber_or_error(sparsity_s_local, 'parameter sparsity_s_local');
check_0leq_param_leq1_or_error(threshold_theta, 'parameter threshold_theta');
check_0leq_param_leq1_or_error(threshold_theta_basic, 'parameter threshold_theta_basic');
check_naturalnumber_or_error(niter_r, 'parameter niter_r');
check_naturalnumber_or_error(niter_r_basic, 'parameter niter_r_basic');

check_mex_file_or_warn('lfft_gen_lattice_mex');
check_mex_file_or_warn('lfft_lattice_size_reduce_bisect_only_mex');
check_mex_file_or_warn('lfft_lattice_size_reduce_mex');

fprintf('sfft: d = %d, N = %d, algorithm = %s, sparsity_s = %d, sparsity_s_local = %d\n', d, N, algorithm_name, sparsity_s, sparsity_s_local);
fprintf('sfft: threshold_theta = %.3e, basic_threshold_theta = %.3e, niter_r = %d, niter_r_basic = %d\n', threshold_theta, threshold_theta_basic, niter_r, niter_r_basic);
%drawnow;

freq_cand_sizes = zeros(1,d);
lattice_sizes = zeros(1,d);
sample_sizes = zeros(1,d);
run_times = zeros(1,5); % basic, lattice search, sampling, lfft, all

freq_cand_sizes(1) = 2*N+1;
lattice_sizes(1) = 2*N+1;

tAll = tic;

if d == 1
  tBasic = tic;
  [ freq, val ] = reco_cardI_basic1( fct_handle, d, N, 1, threshold_theta_basic);
  run_times(1) = toc(tBasic);
  sample_sizes(1) = 2*N+1;
else
  freq = zeros(1,1);
  z = zeros(1,d);
  
  % step 1
  L_1 = 2*N+1;
  for vv=1:niter_r_basic
    tBasic = tic;
    [ freq_local, ~ ] = reco_cardI_basic1( fct_handle, d, N, 1, threshold_theta_basic);
    run_times(1) = run_times(1) + toc(tBasic);
    sample_sizes(1) = sample_sizes(1) + 2*N+1;
    freq = union(freq,freq_local,'rows');
  end
  for S_1=1:L_1
    if length(unique(mod(freq,S_1))) == length(freq)
      break;
    end
  end

  fprintf('step 1: L_1 = %d, card(I^(1)) = %d, S_1 = %d\n', L_1, size(freq,1), S_1);
  
  if S_1 == 1
    z(1) = 0;
  else
    z(1) = 1;
  end
  M = S_1;
  % step 2
  for t = 2:d
    fprintf('step 2: t = %d start\n', t);
    freq_t = zeros(1,1);
    % step 2a
    L_t = 2*N+1;
    for vv=1:niter_r_basic
      tBasic = tic;
      [ freqnew_local, ~ ] = reco_cardI_basic1( fct_handle, d, N, t, threshold_theta_basic);
      run_times(1) = run_times(1) + toc(tBasic);
      sample_sizes(t) = sample_sizes(t) + 2*N+1;
      freq_t = union(freq_t,freqnew_local,'rows');
    end
    fprintf('step 2a: L_%d = %d, card(I^(%d)) = %d\n', t, L_t, t, size(freq_t,1));
    A=reshape(repmat(1:size(freq,1), size(freq_t,1), 1), size(freq,1)*size(freq_t,1), 1);
    B=repmat((1:size(freq_t,1))', size(freq,1), 1);
    indcross = [A B];
    clear A B;
    freq_cand = [freq(indcross(:,1),:) freq_t(indcross(:,2),:)];
    clear indcross;
    if strcmp(algorithm_name, 'a1r1lsymhc')
      freq_cand = freq_cand(prod(max(1,abs(freq_cand)),2) <= N,:);
    end

    freq_cand_sizes(t) = size(freq_cand,1);

    % step 2b
    if t==d
      niter_r_tilde=1;
    else
      niter_r_tilde=niter_r;
    end

    for S_t=1:L_t
      if length(unique(mod(freq_t,S_t))) == length(freq_t)
        break;
      end
    end

    fprintf('step 2b: r_tilde = %d, S_%d = %d\n', niter_r_tilde, t, S_t);

    if (strcmp(algorithm_name, 'a1r1l')) || (strcmp(algorithm_name, 'a1r1lsymhc'))
      fprintf('step 2b: start building reconstructing rank-1 lattice\n');
      tLattice = tic;
      optsSearch.Display='off';
      optsSearch.Inc=0;
      if nthreads > 0
        optsSearch.nthreads = nthreads;
      end
      Mcube = prod(max(freq_cand,[],1)-min(freq_cand,[],1)+1);
      Min = min(M * S_t, Mcube);
      if exist(strcat('lfft_gen_lattice_mex.',mexext), 'file') == 3
        z_out = [];
        try
          z_out = lfft_gen_lattice_mex(freq_cand,Min,optsSearch,z(1:t-1));
        catch
        end
        if length(z_out) ~= t
          warning('z_out too short, using z(t) = M, newM = M * S_t');
          z_out = [z(1:t-1), M];
          Min = M * S_t;
        end
      else
        z_out = [];
        try
          z_out = lfft_gen_lattice(freq_cand,Min,optsSearch,z(1:t-1));
        catch
        end
        if length(z_out) ~= t
          warning('z_out too short, using z(t) = M, newM = M * S_t');
          z_out = [z(1:t-1), M];
          Min = M * S_t;
        end
      end
      if length(z_out) ~= t
        error('length of z');
      end
      z(1:t) = z_out;
      if length(unique(mod(freq_cand*z(1:t)',Min))) ~= size(freq_cand,1)
        error('Invalid generating vector');
      end

      optPmin.Display='off';
      optPmin.Prime=0;
      optPmin.Bisection_Successors=1000;

      if nthreads > 0
        optPmin.nthreads = nthreads;
      end

      P_in=Min;
      if exist(strcat('lfft_lattice_size_reduce_bisect_only_mex.',mexext), 'file') == 3
        P_out=lfft_lattice_size_reduce_bisect_only_mex(freq_cand,P_in,z(1:t),optPmin);
      else
        P_out = lfft_lattice_size_reduce_bisect_only(freq_cand,P_in,z(1:t),optPmin);
      end
      fprintf('step 2b:   np start, M_in = %d, M_out = %d, improvement abs = %d / rel = %.3f\n', P_in, P_out, P_in-P_out, (P_in-P_out)/P_in);
      P_in=P_out;

      optPmin.Display='off';
      optPmin.Prime=1;
      optPmin.Bisection_Successors=1000;
      for r=1:10
        if exist(strcat('lfft_lattice_size_reduce_bisect_only_mex.',mexext), 'file') == 3
          P_out=lfft_lattice_size_reduce_bisect_only_mex(freq_cand,P_in,z(1:t),optPmin);
        else
          P_out = lfft_lattice_size_reduce_bisect_only(freq_cand,P_in,z(1:t),optPmin);
        end
        fprintf('step 2b:    p r =%2d, M_in = %d, M_out = %d, improvement abs = %d / rel = %.3f\n', r, P_in, P_out, P_in-P_out, (P_in-P_out)/P_in);
        if P_in==P_out
          break;
        end
        P_in = P_out;
      end

      M=P_out;
      run_times(2) = run_times(2) + toc(tLattice);
      fprintf('step 2b: finished building reconstructing rank-1 lattice\n');
    else
      z(t) = M;
      M = M * S_t;
    end

    fprintf('step 2b: z_%d = %d, z(1:%d) = [', t, z(t), t);
    for s=1:t
      fprintf(' %d', z(s));
    end
    fprintf(' ], M = %d\n', M);

    if length(unique(mod(freq_cand*z(1:t)',M))) ~= size(freq_cand,1)
      error('rank-1 lattice does not fulfill reconstructing property');
    end

    lattice_sizes(t) = M;

    fprintf('step 2c: %d samples, r_tilde = %d detection iterations\n', M, niter_r_tilde);
    %drawnow;

    ind = [];
    for vv=1:niter_r_tilde
      % step 2c
      tSampling = tic;
      try
        samples = fct_handle(z(1:t),M,rand(d-t,1));
      catch
        xj = repmat(rand(1,d),M,1);
        xj(:,1:t) = mod((0:M-1)' * z(1:t), M)/M;
        samples = fct_handle(xj);
        clear xj;
      end
      run_times(3) = run_times(3) + toc(tSampling);
      sample_sizes(t) = sample_sizes(t) + M;

      % step 2d
      tLFFT = tic;
      val_all = lfft(freq_cand,z(1:t),M,samples,'transp')/M;
      run_times(4) = run_times(4) + toc(tLFFT);
      clear samples;

      % step 2e
      max_abs_val_all = max(abs(val_all));
      absolute_threshold = threshold_theta*max_abs_val_all;
      ind_local = find(abs(val_all) >= absolute_threshold);
      clear max_abs_val_all;
      if length(ind_local) > sparsity_s_local
        [~,ind_local] = sort(abs(val_all),'descend');
        ind_local = ind_local(1:sparsity_s_local);
      end
      fprintf('step 2e: i = %d, max fc_all = %.1e, min fc_all = %.1e, max fc_local = %.1e, min fc_local = %.1e\n', vv, max(abs(val_all)), min(abs(val_all)), max(abs(val_all(ind_local))), min(abs(val_all(ind_local))));

      if vv==1
        ind = ind_local;
      else
        ind = union(ind,ind_local,'rows');
      end
      clear ind_local;

      fprintf('step 2e: i = %d, card(I^(1,...,%d)) = %d\n', vv, t, length(ind));
      %drawnow;
    end
    freq = freq_cand(ind,:);
    val = val_all(ind);
    if ((t==d) && (length(val) > sparsity_s))
      [~,ind] = sort(abs(val),'descend');
      val = val(ind(1:sparsity_s));
      freq = freq(ind(1:sparsity_s),:);
    end
    fprintf('step 2e: result dim %d of %d: %d freq candidates, %d samples, %d freq detected\n', t, d, size(freq_cand,1), M, size(freq,1));
    %drawnow;
    clear ind;

    % additional step 2f
    if t ~= d
      tLattice = tic;
      freq_lengths = max(freq,[],1)-min(freq,[],1)+1;
      Mcube = prod(freq_lengths);
      Min = min(M, Mcube);
      optsSearch.Display='off';
      optsSearch.Inc=0;
      if nthreads > 0
        optsSearch.nthreads = nthreads;
      end
      if Min < M
        z_out = ones(1,t);
        for j = 2:t
          z_out(j) = z_out(j-1) * freq_lengths(j-1);
        end
      else
        if exist(strcat('lfft_gen_lattice_mex.',mexext), 'file') == 3
          z_out = lfft_gen_lattice_mex(freq,Min,optsSearch,z(1:t-1));
        else
          z_out = lfft_gen_lattice(freq,Min,optsSearch,z(1:t-1));
        end
      end
      if length(z_out) ~= t
        error('length of z');
      end
      z(1:t) = z_out;
      if length(unique(mod(freq*z(1:t)',Min))) ~= size(freq,1)
        error('Invalid generating vector');
      end
      
      optsReduce.Display='off';
      optsReduce.Prime=1;
      optsReduce.Bisection_Successors=1000;
      if t==2
        optsReduce.Mmin = size(freq,1);
      else
        optsReduce.Mmin = min(size(freq,1),lattice_sizes(t-1));
      end
      if nthreads > 0
        optsReduce.nthreads = nthreads;
      end
      if exist(strcat('lfft_lattice_size_reduce_mex.',mexext), 'file') == 3
        M = lfft_lattice_size_reduce_mex(freq,Min,z(1:t),optsReduce);
      else
        M = lfft_lattice_size_reduce(freq,Min,z(1:t),optsReduce);
      end
      run_times(2) = run_times(2) + toc(tLattice);

      fprintf('additional step 2f: z_%d = %d, z(1:%d) = [', t, z(t), t);
      for s=1:t
        fprintf(' %d', z(s));
      end
      fprintf(' ], M_%d = %d\n', t, M);

      if length(unique(mod(freq*z(1:t)',M))) ~= size(freq,1)
        error('rank-1 lattice does not fulfill reconstructing property');
      end
    end
    fprintf('step 2: t = %d end, d = %d, M = %d\n\n', t, d, M);
    %drawnow;
  end
  fprintf('step 3: %.10g possible freq., %d total samples, %d freq. detected\n', (2*N+1)^d, sum(sample_sizes), size(freq,1));
  %drawnow;
end

run_times(end) = toc(tAll);

end

function [val] = check_naturalnumber_or_error(param, name)
  val = 0;
  if (param < 1) || (param ~= int32(param))
    error('%s must be a natural number', name);
  end
end

function [val] = check_0leq_param_leq1_or_error(param, name)
  val = 0;
  if (param < 0) || (param > 1)
    error('%s must be >= 0 and <= 1', name);
  end
end

function [filename] = check_mex_file_or_warn(filename_without_ext)
  filename = strcat(filename_without_ext,strcat('.',mexext));
  if exist(filename, 'file') ~= 3
    warning('MATLAB mex file %s is missing. The algorithm will be very slow!\nPlease use the MATLAB script compile_mex_files_without_openmp.m or compile_mex_files_with_openmp.m', filename);
  end
end
